// $Id: tideDialog.cpp 4046 2011-09-07 21:43:34Z flaterco $

/*  TideDialog class.  */

/*****************************************************************************\

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\*****************************************************************************/



#define __STDC_FORMAT_MACROS
#include "tideDialog.h"
#include "tideDialogHelp.h"
#include <inttypes.h>
#include <strings.h>
#include <assert.h>
#include <QDesktopWidget>  // 20110905
//Added by qt3to4:
#include <QLabel>
#include <QPixmap>
#include <Q3Frame>


//  This is the alpha sort function for qsort.

static int compare_names (const void *a, const void *b)
{
    return strcasecmp (((CONSTITUENT*)a)->name, ((CONSTITUENT*)b)->name);
}


static void safeStrCpy (char *to, const char *from, unsigned bufsize) {
  assert (from);
  assert (to);
  assert (bufsize > 0);
  strncpy (to, from, bufsize-1);
  to[bufsize-1] = '\0';
}


static void safeStrCpy (char *to, const QString &from, unsigned bufsize) {
  safeStrCpy (to, from.latin1(), bufsize);
}


static void safeStrCpy (char *to, const QLineEdit &from, unsigned bufsize) {
  safeStrCpy (to, from.text(), bufsize);
}


static void safeStrCpy (char *to, const Q3TextEdit &from, unsigned bufsize) {
  safeStrCpy (to, from.text(), bufsize);
}


//  If the TIDE_RECORD, lrec, is entered with a header.record_number less than
//  zero we will be adding a new record.

TideDialog::TideDialog (QWidget *parent, const char *name, TIDE_RECORD lrec, NV_BOOL ro, 
                        TideEditor *te):
  Q3TabDialog (parent, name)
{
    unsigned       i;


    readonly = ro;


    db = get_tide_db_header ();
    rec = lrec;
    tideEditor = te;

    sub_station = NULL;


    cnst = (CONSTITUENT *) calloc (db.constituents, sizeof (CONSTITUENT));
    if (cnst == NULL)
      {
        perror ("Allocating constituent list");
        exit (-1);
      }


    for (i = 0 ; i < db.constituents ; ++i)
    {
        cnst[i].num = i;
        safeStrCpy (cnst[i].name, get_constituent (i), ConstituentNameLen);
    }


    //  Sort the names then update the constituent list.

    ::qsort (cnst, db.constituents, sizeof (CONSTITUENT), &compare_names);


    tideDialog = new Q3TabDialog (this, 0, TRUE, Qt::WDestructiveClose);
    tideDialog->resize (640, 640);
    tideDialog->setCaption (name);
    tideDialog->setCancelButton ("Cancel");
    tideDialog->setHelpButton ("Print");

    connect (tideDialog, SIGNAL (applyButtonPressed ()), this, 
        SLOT (slotUpdateStation ()));
    connect (tideDialog, SIGNAL (helpButtonPressed ()), this, 
        SLOT (slotPrintStation ()));
    connect (tideDialog, SIGNAL (cancelButtonPressed ()), this, 
        SLOT (slotCancelStation ()));


    if (rec.header.record_number >= 0)
    {
        tideDialog->setDefaultButton ("Delete Record");
        connect (tideDialog, SIGNAL (defaultButtonPressed ()), this, 
            SLOT (slotDeleteStation ()));
    }


    //  Build the normal tabs.

    generalTab ();
    constituentsTab ();
    offsetsTab ();

    Q3WhatsThis::add (tideDialog, tideDialogText);

    if (QApplication::desktop ()->width () > (640 + 50)
        && QApplication::desktop ()->height () > (640 + 66))
    {
        tideDialog->show ();
    }
    else
    {
        tideDialog->showMaximized ();
    }
}




TideDialog::~TideDialog ()
{
    if (cnst) free (cnst);
}



//  Information that is common to both Reference and Subordinate stations.

void
TideDialog::generalTab ()
{
    char hem;
    unsigned i;
    double deg, min, sec;

    Q3VBox *generalTab = new Q3VBox (this);
    generalTab->setMargin (5);
    generalTab->setSpacing (5);

    QString rcnm = QString ("%1").arg (rec.header.record_number);
    if (rec.header.record_number == -1) rcnm = "";
    QLabel *reclab = new QLabel ("Record Number:", generalTab);
    QLabel *recnum = new QLabel (rcnm, generalTab);
    recnum->setFrameStyle (Q3Frame::Panel | Q3Frame::Sunken);
    Q3WhatsThis::add (reclab, recnumText);
    Q3WhatsThis::add (recnum, recnumText);

    if (rec.header.reference_station >= 0)
    {
        QString rfnm (get_station (rec.header.reference_station));
        QLabel *reflab = new QLabel ("Reference Station:", generalTab);
        QLabel *ref = new QLabel (rfnm, generalTab);
        ref->setFrameStyle (Q3Frame::Panel | Q3Frame::Sunken);
        Q3WhatsThis::add (reflab, refText);
        Q3WhatsThis::add (ref, refText);
    }

    QString stnm (rec.header.name);
    QLabel *namelab = new QLabel ("Station Name:", generalTab);
    name = new QLineEdit (stnm, generalTab);
    name->setFocus ();
    Q3WhatsThis::add (namelab, nameText);
    Q3WhatsThis::add (name, nameText);

    {
      QLabel *label = new QLabel ("Station ID Context:", generalTab);
      station_id_context = new QLineEdit (QString (rec.station_id_context), generalTab);
      Q3WhatsThis::add (label, station_id_contextText);
      Q3WhatsThis::add (station_id_context, station_id_contextText);
    }{
      QLabel *label = new QLabel ("Station ID:", generalTab);
      station_id = new QLineEdit (QString (rec.station_id), generalTab);
      Q3WhatsThis::add (label, station_idText);
      Q3WhatsThis::add (station_id, station_idText);
    }

    QString lats = QString (fixpos (rec.header.latitude, &deg, &min, &sec, 
        &hem, "lat", tideEditor->pos_format));
    QLabel *latlab = new QLabel ("Latitude:", generalTab);
    lat = new QLineEdit (lats, generalTab);
    Q3WhatsThis::add (latlab, latText);
    Q3WhatsThis::add (lat, latText);

    QString lons = QString (fixpos (rec.header.longitude, &deg, &min, &sec, 
        &hem, "lon", tideEditor->pos_format));
    QLabel *lonlab = new QLabel ("Longitude:", generalTab);
    lon = new QLineEdit (lons, generalTab);
    Q3WhatsThis::add (lonlab, lonText);
    Q3WhatsThis::add (lon, lonText);

    QLabel *tzfiles = new QLabel ("Time Zone:", generalTab);
    tzfile = new QComboBox (FALSE, generalTab);
    Q3WhatsThis::add (tzfiles, tzfileText);
    Q3WhatsThis::add (tzfile, tzfileText);
    for (i = 0 ; i < db.tzfiles ; ++i) 
    {
	QString str = QString (get_tzfile (i));
        tzfile->insertItem (str);
    }
    tzfile->setCurrentItem (rec.header.tzfile);

    QLabel *countrys = new QLabel ("Country:", generalTab);
    country = new QComboBox (FALSE, generalTab);
    Q3WhatsThis::add (countrys, countryText);
    Q3WhatsThis::add (country, countryText);
    for (i = 0 ; i < db.countries ; ++i) 
    {
	QString str = QString (get_country (i));
        country->insertItem (str);
    }
    country->setCurrentItem (rec.country);

    QLabel *direction_unitss = new QLabel ("Direction Units:", generalTab);
    direction_units = new QComboBox (FALSE, generalTab);
    Q3WhatsThis::add (direction_unitss, direction_unitsText);
    Q3WhatsThis::add (direction_units, direction_unitsText);
    for (i = 0 ; i < db.dir_unit_types ; ++i) 
    {
	QString str = QString (get_dir_units (i));
        direction_units->insertItem (str);
    }
    direction_units->setCurrentItem (rec.direction_units);

    QString min_directions = QString ("%1").arg (rec.min_direction);
    if (min_directions == "361") min_directions = "";
    QLabel *min_directionlab = new QLabel ("Ebb Direction:", generalTab);
    min_direction = new QLineEdit (min_directions, generalTab);
    Q3WhatsThis::add (min_directionlab, min_directionText);
    Q3WhatsThis::add (min_direction, min_directionText);

    QString max_directions = QString ("%1").arg (rec.max_direction);
    if (max_directions == "361") max_directions = "";
    QLabel *max_directionlab = new QLabel ("Flood Direction:", generalTab);
    max_direction = new QLineEdit (max_directions, generalTab);
    Q3WhatsThis::add (max_directionlab, max_directionText);
    Q3WhatsThis::add (max_direction, max_directionText);

    QLabel *level_unitss = new QLabel ("Level Units:", generalTab);
    level_units = new QComboBox (FALSE, generalTab);
    Q3WhatsThis::add (level_unitss, level_unitsText);
    Q3WhatsThis::add (level_units, level_unitsText);
    for (i = 0 ; i < db.level_unit_types ; ++i) 
    {
	QString str = QString (get_level_units (i));
        level_units->insertItem (str);
    }
    level_units->setCurrentItem (rec.level_units);

    (void) Q3WhatsThis::whatsThisButton (generalTab);
    tideDialog->addTab (generalTab, "General");


    // General pt. 2 (verbiage)

    generalTab = new Q3VBox (this);
    generalTab->setMargin (5);
    generalTab->setSpacing (5);

    {
      QLabel *label = new QLabel ("Date Imported:", generalTab);
      QString date_importeds = QString ("%1").arg (rec.date_imported);
      if (date_importeds == "0") date_importeds = "";
      date_imported = new QLineEdit (date_importeds, generalTab);
      Q3WhatsThis::add (label, date_importedText);
      Q3WhatsThis::add (date_imported, date_importedText);
    }

    QString src = QString (rec.source);
    QLabel *sourcelab = new QLabel ("Source:", generalTab);
    source = new QLineEdit (src, generalTab);
    Q3WhatsThis::add (sourcelab, sourceText);
    Q3WhatsThis::add (source, sourceText);

    QLabel *restrictions = new QLabel ("Restriction:", generalTab);
    restriction = new QComboBox (FALSE, generalTab);
    Q3WhatsThis::add (restrictions, restrictionText);
    Q3WhatsThis::add (restriction, restrictionText);
    for (i = 0 ; i < db.restriction_types ; ++i) 
    {
	QString str = QString (get_restriction (i));
        restriction->insertItem (str);
    }
    restriction->setCurrentItem (rec.restriction);

    QLabel *commentslab = new QLabel ("Comments:", generalTab);
    comments = new Q3TextEdit (generalTab);
    comments->setTextFormat (Qt::PlainText);
    comments->setText (QString (rec.comments));
    Q3WhatsThis::add (commentslab, commentsText);
    Q3WhatsThis::add (comments, commentsText);

    QLabel *noteslab = new QLabel ("Notes:", generalTab);
    notes = new Q3TextEdit (generalTab);
    notes->setTextFormat (Qt::PlainText);
    notes->setText (QString (rec.notes));
    Q3WhatsThis::add (noteslab, notesText);
    Q3WhatsThis::add (notes, notesText);

    QLabel *legaleses = new QLabel ("Legalese:", generalTab);
    legalese = new QComboBox (FALSE, generalTab);
    Q3WhatsThis::add (legaleses, legaleseText);
    Q3WhatsThis::add (legalese, legaleseText);
    for (i = 0 ; i < db.legaleses ; ++i) 
    {
	QString str = QString (get_legalese (i));
        legalese->insertItem (str);
    }
    legalese->setCurrentItem (rec.legalese);

    QLabel *xfieldslab = new QLabel ("X-fields:", generalTab);
    xfields = new Q3TextEdit (generalTab);
    xfields->setTextFormat (Qt::PlainText);
    xfields->setText (QString (rec.xfields));
    Q3WhatsThis::add (xfieldslab, xfieldsText);
    Q3WhatsThis::add (xfields, xfieldsText);

    (void) Q3WhatsThis::whatsThisButton (generalTab);
    tideDialog->addTab (generalTab, "Verbiage");
}



//  Information that is unique to Reference stations.

void
TideDialog::constituentsTab ()
{
    unsigned i;

    //  Don't display this if this is a Subordinate record.

    if (rec.header.record_type == SUBORDINATE_STATION) return;


    Q3VBox *constituentsTab = new Q3VBox (this);
    constituentsTab->setMargin (5);
    constituentsTab->setSpacing (5);

    QLabel *datums = new QLabel ("Datum Description:", constituentsTab);
    datum = new QComboBox (FALSE, constituentsTab);
    Q3WhatsThis::add (datums, datumText);
    Q3WhatsThis::add (datum, datumText);
    for (i = 0 ; i < db.datum_types ; ++i) 
    {
	QString str = QString (get_datum (i));
        datum->insertItem (str);
    }
    datum->setCurrentItem (rec.datum);

    QString datum_offsets = QString ("%1").arg (rec.datum_offset, 0, 'f', 4);
    QLabel *datum_offsetlab = new QLabel ("Datum Value:", constituentsTab);
    datum_offset = new QLineEdit (datum_offsets, constituentsTab);
    Q3WhatsThis::add (datum_offsetlab, datum_offsetText);
    Q3WhatsThis::add (datum_offset, datum_offsetText);

    QString zone_offsets = QString ("%1").arg (rec.zone_offset);
    QLabel *zone_offsetlab = new QLabel ("Meridian (a.k.a. Zone Offset):", constituentsTab);
    zone_offset = new QLineEdit (zone_offsets, constituentsTab);
    Q3WhatsThis::add (zone_offsetlab, zone_offsetText);
    Q3WhatsThis::add (zone_offset, zone_offsetText);

    QString expiration_dates = QString ("%1").arg (rec.expiration_date);
    if (expiration_dates == "0") expiration_dates = "";
    QLabel *expiration_datelab = 
        new QLabel ("Expiration Date:", constituentsTab);
    expiration_date = new QLineEdit (expiration_dates, constituentsTab);
    Q3WhatsThis::add (expiration_datelab, expiration_dateText);
    Q3WhatsThis::add (expiration_date, expiration_dateText);

    QString monthss = QString ("%1").arg (rec.months_on_station);
    if (monthss == "0") monthss = "";
    QLabel *monthslab = new QLabel ("Months On Station:", constituentsTab);
    months = new QLineEdit (monthss, constituentsTab);
    Q3WhatsThis::add (monthslab, monthsText);
    Q3WhatsThis::add (months, monthsText);

    QString ldoss = QString ("%1").arg (rec.last_date_on_station);
    if (ldoss == "0") ldoss = "";
    QLabel *ldoslab = new QLabel ("Last Date On Station:", constituentsTab);
    ldos = new QLineEdit (ldoss, constituentsTab);
    Q3WhatsThis::add (ldoslab, ldosText);
    Q3WhatsThis::add (ldos, ldosText);

    QString confidences = QString ("%1").arg (rec.confidence);
    QLabel *confidencelab = new QLabel ("Confidence:", constituentsTab);
    confidence = new QLineEdit (confidences, constituentsTab);
    Q3WhatsThis::add (confidencelab, confidenceText);
    Q3WhatsThis::add (confidence, confidenceText);

    (void) new QLabel ("Constituents:", constituentsTab);

    constituents = new Q3Table (db.constituents, 2, constituentsTab);
    Q3WhatsThis::add (constituents, constituentsText);

    Q3Header *hh = constituents->horizontalHeader ();
    hh->setLabel (0, QObject::tr ("Amplitude"), -1);
    hh->setLabel (1, QObject::tr ("Epoch"), -1);

    Q3Header *vh = constituents->verticalHeader ();
    vh->setOrientation (Qt::Vertical);

    inf_rec = rec;
    infer_constituents (&inf_rec);

    QPixmap inferIcon = QPixmap (inferred);

    Q3MimeSourceFactory::defaultFactory ()->setPixmap ("inferred", inferIcon);

    for (i = 0 ; i < db.constituents ; ++i)
    {
        QString cnstnm (cnst[i].name);
        vh->setLabel (i, cnstnm, -1);
        int j = cnst[i].num;

        QString amplitudes, epochs;
        if (rec.amplitude[j] == 0.0 && rec.epoch[j] == 0.0)
        {
            if (inf_rec.amplitude[j] == 0.0 && inf_rec.epoch[j] == 0.0)
            {
                amplitudes = QString ("");
                epochs = QString ("");
            }
            else
            {
                //  Inferred constituents.

                amplitudes = QString ("%1").arg (inf_rec.amplitude[j], 0, 
                    'f', 4);
                epochs = QString ("%1").arg (inf_rec.epoch[j], 0, 'f', 2);

                constituents->setPixmap (i, 0, inferIcon);
                constituents->setPixmap (i, 1, inferIcon);
            }
        }
        else
        {
            inf_rec.amplitude[j] = 0.0;
            inf_rec.epoch[j] = 0.0;


            amplitudes = QString ("%1").arg (rec.amplitude[j], 0, 'f', 4);
            epochs = QString ("%1").arg (rec.epoch[j], 0, 'f', 2);
        }

        constituents->setText (i, 0, amplitudes);
        constituents->setText (i, 1, epochs);


        //  Make sure that we can detect any change in the constituents 
        //  so that we only replace those that have changed.  This way we don't
        //  save inferred constituents unless they have been modified.

        sscanf (constituents->text (i, 0), "%f", &inf_rec.amplitude[j]);
        sscanf (constituents->text (i, 1), "%f", &inf_rec.epoch[j]);
    }


    (void) Q3WhatsThis::whatsThisButton (constituentsTab);

    tideDialog->addTab (constituentsTab, "Constituents");


    //  Check for subordinate stations so we can build a list tab if
    //  there are any.

    if (rec.header.record_number >= 0 && rec.header.reference_station == -1)
    {
        TIDE_RECORD temp_rec;
        unsigned i, subs = 0;
        
        Q3VBox *subTab = new Q3VBox (this);
        subTab->setMargin (5);
        subTab->setSpacing (5);

        subList = new Q3ListBox (subTab);
        Q3WhatsThis::add (subList, subListText);
        subList->setSelectionMode (Q3ListBox::Extended);


        if (sub_station) free (sub_station);


        for (i = 0 ; i < db.number_of_records ; ++i)
        {
            get_partial_tide_record (i, &temp_rec.header);

            if (temp_rec.header.reference_station == 
                rec.header.record_number)
            {
                double deg, min, sec;
                char hem;

                if ((sub_station = (int *) realloc (sub_station, (subs + 1) * 
                    sizeof (int))) == NULL)
                {
                    perror ("Allocating sub station memory");
                    exit (-1);
                }

                QString lats = QString (fixpos (temp_rec.header.latitude, 
                    &deg, &min, &sec, &hem, "lat", tideEditor->pos_format));
                QString lons = QString (fixpos (temp_rec.header.longitude, 
                    &deg, &min, &sec, &hem, "lon", tideEditor->pos_format));

                QString line = QString ("%1, %2   : %3")
                    .arg (lats, 13)
                    .arg (lons, 14)
                    .arg (temp_rec.header.name);

                subList->insertItem (line, -1);

                sub_station[subs] = temp_rec.header.record_number;
                ++subs;
            }
        }


        //  If there were any, build the tab.

        if (subs)
        {
            QPushButton *changeButton = 
                new QPushButton ("Change Reference Station", subTab, 0);
            Q3WhatsThis::add (changeButton, subListText);

            connect (changeButton, SIGNAL (clicked ()), this, 
                SLOT (slotChangeRefStation ()));


            (void) Q3WhatsThis::whatsThisButton (subTab);

            tideDialog->addTab (subTab, "Subordinate Stations");
        }
        else
        {
            delete (subList);
        }
    }
}



//  Information that is unique to Subordinate stations.

void
TideDialog::offsetsTab ()
{
    //  If this is a Reference station, don't display this tab.

    if (rec.header.record_type == REFERENCE_STATION) return;

    Q3VBox *offsetsTab = new Q3VBox (this);
    offsetsTab->setMargin (5);
    offsetsTab->setSpacing (5);

    QString min_time_adds = QString ("%1").arg (rec.min_time_add);
    if (min_time_adds == "0") min_time_adds = "";
    QLabel *min_time_addlab = new QLabel ("Minimum Time Add:", offsetsTab);
    min_time_add = new QLineEdit (min_time_adds, offsetsTab);
    Q3WhatsThis::add (min_time_addlab, min_time_addText);
    Q3WhatsThis::add (min_time_add, min_time_addText);

    QString min_level_adds = QString ("%1").arg (rec.min_level_add, 0, 'f', 
        2);
    if (min_level_adds == "0.00") min_level_adds = "";
    QLabel *min_level_addlab = new QLabel ("Minimum Level Add:", offsetsTab);
    min_level_add = new QLineEdit (min_level_adds, offsetsTab);
    Q3WhatsThis::add (min_level_addlab, min_level_addText);
    Q3WhatsThis::add (min_level_add, min_level_addText);

    QString min_level_mults = QString ("%1").arg (rec.min_level_multiply, 0, 
        'f', 2);
    if (min_level_mults == "0.00") min_level_mults = "";
    QLabel *min_level_multlab = 
        new QLabel ("Minimum Level Multiply:", offsetsTab);
    min_level_mult = new QLineEdit (min_level_mults, offsetsTab);
    Q3WhatsThis::add (min_level_multlab, min_level_multText);
    Q3WhatsThis::add (min_level_mult, min_level_multText);

    QString max_time_adds = QString ("%1").arg (rec.max_time_add);
    if (max_time_adds == "0") max_time_adds = "";
    QLabel *max_time_addlab = new QLabel ("Maximum Time Add:", offsetsTab);
    max_time_add = new QLineEdit (max_time_adds, offsetsTab);
    Q3WhatsThis::add (max_time_addlab, max_time_addText);
    Q3WhatsThis::add (max_time_add, max_time_addText);

    QString max_level_adds = QString ("%1").arg (rec.max_level_add, 0, 'f', 
        2);
    if (max_level_adds == "0.00") max_level_adds = "";
    QLabel *max_level_addlab = new QLabel ("Maximum Level Add:", offsetsTab);
    max_level_add = new QLineEdit (max_level_adds, offsetsTab);
    Q3WhatsThis::add (max_level_addlab, max_level_addText);
    Q3WhatsThis::add (max_level_add, max_level_addText);

    QString max_level_mults = QString ("%1").arg (rec.max_level_multiply, 0,
        'f', 2);
    if (max_level_mults == "0.00") max_level_mults = "";
    QLabel *max_level_multlab = 
        new QLabel ("Maximum Level Multiply:", offsetsTab);
    max_level_mult = new QLineEdit (max_level_mults, offsetsTab);
    Q3WhatsThis::add (max_level_multlab, max_level_multText);
    Q3WhatsThis::add (max_level_mult, max_level_multText);

    /* DWF updated per slack offsets fix 2003-03-18 */
    QString flood_beginss = QString ("%1").arg (rec.flood_begins);
    if (rec.flood_begins == NULLSLACKOFFSET) flood_beginss = "";
    QLabel *flood_beginslab = new QLabel ("Flood Begins:", offsetsTab);
    flood_begins = new QLineEdit (flood_beginss, offsetsTab);
    Q3WhatsThis::add (flood_beginslab, flood_beginsText);
    Q3WhatsThis::add (flood_begins, flood_beginsText);

    /* DWF updated per slack offsets fix 2003-03-18 */
    QString ebb_beginss = QString ("%1").arg (rec.ebb_begins);
    if (rec.ebb_begins == NULLSLACKOFFSET) ebb_beginss = "";
    QLabel *ebb_beginslab = new QLabel ("Ebb Begins:", offsetsTab);
    ebb_begins = new QLineEdit (ebb_beginss, offsetsTab);
    Q3WhatsThis::add (ebb_beginslab, ebb_beginsText);
    Q3WhatsThis::add (ebb_begins, ebb_beginsText);

    (void) Q3WhatsThis::whatsThisButton (offsetsTab);

    tideDialog->addTab (offsetsTab, "Offsets");
}


//  Update station data in tide DB.

void
TideDialog::slotUpdateStation ()
{
    if (readonly)
      {
        QMessageBox::information (0, tr ("tideDialog"), 
                                  tr ("The database is opened in read-only mode.  No changes are allowed."));

          return;
      }


    unsigned i;
    double inlat, inlon;
    static const unsigned coordStringLength = 50U;
    char string[coordStringLength];


    // Common data.

    safeStrCpy (rec.header.name, *name, ONELINER_LENGTH);
    safeStrCpy (rec.source, *source, ONELINER_LENGTH);
    safeStrCpy (rec.station_id_context, *station_id_context, ONELINER_LENGTH);
    safeStrCpy (rec.station_id, *station_id, ONELINER_LENGTH);
    if (sscanf (date_imported->text (), "%u", &rec.date_imported) <= 0)
      rec.date_imported = 0;

    safeStrCpy (string, *lat, coordStringLength);
    posfix (string, &inlat, "lat");
    safeStrCpy (string, *lon, coordStringLength);
    posfix (string, &inlon, "lon");

    if (inlat != rec.header.latitude || inlon != rec.header.longitude)
    {
        rec.header.latitude = inlat;
        rec.header.longitude = inlon;
        tideEditor->setRedrawZeroLevel ();
    }


    rec.header.tzfile = tzfile->currentItem ();
    rec.country = country->currentItem ();
    rec.restriction = restriction->currentItem ();
    safeStrCpy (rec.comments, *comments, MONOLOGUE_LENGTH);
    safeStrCpy (rec.notes, *notes, MONOLOGUE_LENGTH);
    safeStrCpy (rec.xfields, *xfields, MONOLOGUE_LENGTH);
    rec.direction_units = direction_units->currentItem ();
    if (sscanf (min_direction->text (), "%d", &rec.min_direction) <= 0)
	rec.min_direction = 361;
    if (sscanf (max_direction->text (), "%d", &rec.max_direction) <= 0)
	rec.max_direction = 361;
    rec.level_units = level_units->currentItem ();
    rec.legalese = legalese->currentItem ();


    //  Reference station data.

    if (rec.header.record_type == REFERENCE_STATION)
    {
        rec.datum = datum->currentItem ();
        sscanf (datum_offset->text (), "%f", &rec.datum_offset);
        sscanf (zone_offset->text (), "%d", &rec.zone_offset);
        if (sscanf (expiration_date->text (), "%u", &rec.expiration_date) <= 0)
            rec.expiration_date = 0;
        if (sscanf (months->text (), "%" SCNu16, &rec.months_on_station) <= 0)
            rec.months_on_station = 0;
        if (sscanf (ldos->text (), "%u", &rec.last_date_on_station) <= 0)
            rec.last_date_on_station = 0;
        if (sscanf (confidence->text (), "%" SCNu8, &rec.confidence) <= 0)
  	    rec.confidence = 0;

        for (i = 0 ; i < db.constituents ; ++i)
        {
            float amp, epo;
            if (sscanf (constituents->text (i, 0), "%f", &amp) != 1) 
                amp = 0.0;
            if (sscanf (constituents->text (i, 1), "%f", &epo) != 1) 
                epo = 0.0;

            int j = cnst[i].num;

            //  Check for a change to constituents.  If there is no
            //  change, don't replace the constituents.

            if (amp != inf_rec.amplitude[j] || epo != inf_rec.epoch[j])
            {
                rec.amplitude[j] = amp;
                rec.epoch[j] = epo;
            }
        }
    }


    //  Subordinate station data.

    else
    {
        if (sscanf (min_time_add->text (), "%d", &rec.min_time_add) <= 0)
            rec.min_time_add = 0;
        if (sscanf (min_level_add->text (), "%f", &rec.min_level_add) <= 0)
            rec.min_level_add = 0;
        if (sscanf (min_level_mult->text (), "%f", &rec.min_level_multiply) <=
            0) rec.min_level_multiply = 0;
        if (sscanf (max_time_add->text (), "%d", &rec.max_time_add) <= 0)
            rec.max_time_add = 0;
        if (sscanf (max_level_add->text (), "%f", &rec.max_level_add) <= 0)
            rec.max_level_add = 0;
        if (sscanf (max_level_mult->text (), "%f", &rec.max_level_multiply) <=
            0) rec.max_level_multiply = 0;
        /* DWF updated per slack offsets fix 2003-03-18 */
        if (sscanf (flood_begins->text (), "%d", &rec.flood_begins) <= 0)
            rec.flood_begins = NULLSLACKOFFSET;
        /* DWF updated per slack offsets fix 2003-03-18 */
        if (sscanf (ebb_begins->text (), "%d", &rec.ebb_begins) <= 0)
            rec.ebb_begins = NULLSLACKOFFSET;
    }


    //  Delete stuff.
    slotCancelStation ();


    //  Update the record in the database, add it if the header.record_number
    //  is less than zero.

    if (rec.header.record_number >= 0)
    {
	if (!(update_tide_record (rec.header.record_number, &rec, &db)))
	  QMessageBox::information (0, tr ("tideDialog"),
	  tr ("Update rejected.  See stderr for details."));
    }
    else
    {
	if (!(add_tide_record (&rec, &db)))
	  QMessageBox::information (0, tr ("tideDialog"),
	  tr ("New record rejected.  See stderr for details."));

        tideEditor->setRedrawZeroLevel ();
    }


    tideEditor->redrawMap ();
}


// Delete stuff.  Called by slotUpdateStation, also called on cancel.

// DWF 2007-04:  Memory management is inconsistent and unclear.  Lots
// of things are deleted here, but subList isn't and labels aren't.
// My reading of the Qt documentation suggests that widgets are
// deleted automatically when their parents are.  Who knows?

void
TideDialog::slotCancelStation ()
{
    if (rec.header.record_type == REFERENCE_STATION)
    {
        delete (datum);
        delete (datum_offset);
        delete (zone_offset);
        delete (expiration_date);
        delete (months);
        delete (ldos);
        delete (confidence);
        delete (constituents);
    }
    else
    {
        delete (min_time_add);
        delete (min_level_add);
        delete (min_level_mult);
        delete (max_time_add);
        delete (max_level_add);
        delete (max_level_mult);
        delete (flood_begins);
        delete (ebb_begins);
    }


    delete (name);
    delete (source);
    delete (station_id_context);
    delete (station_id);
    delete (date_imported);
    delete (lat);
    delete (lon);
    delete (tzfile);
    delete (country);
    delete (restriction);
    delete (comments);
    delete (notes);
    delete (xfields);
    delete (direction_units);
    delete (min_direction);
    delete (max_direction);
    delete (level_units);
    delete (legalese);
    delete (tideDialog);
}



//  Deleting a station.

void 
TideDialog::slotDeleteStation ()
{
    if (readonly)
      {
        QMessageBox::information (0, tr ("tideDialog"), 
                                  tr ("The database is opened in read-only mode.  No changes are allowed."));

          return;
      }


    //  Clear memory.

    slotCancelStation ();


    QMessageBox mb ("Delete Tide Station",
        "Do you really want to delete this station?\n"
        "There is no way to recover once it is deleted!\n\n",
        QMessageBox::Warning,
        QMessageBox::Yes | QMessageBox::Escape,
        QMessageBox::No  | QMessageBox::Default,
        0);

    if (mb.exec() == QMessageBox::No) return;


    //  Check for subordinate stations.

    if (rec.header.record_type == REFERENCE_STATION &&
        rec.header.reference_station == -1) // ?
    {
        TIDE_RECORD temp_rec;
        unsigned i, subs = 0;
        
        for (i = 0 ; i < db.number_of_records ; ++i)
        {
            get_partial_tide_record (i, &temp_rec.header);

            if (temp_rec.header.reference_station == 
                rec.header.record_number) ++subs;
        }

        if (subs)
        {
            QString msg = 
                QString ("This station has %1 subordinate stations!\n"
                "Deleting it will delete all of the subordinate stations.\n"
                "Are you sure you want to delete it?\n\n")
                .arg (subs);

            QMessageBox mb ("Delete Reference Tide Station", msg,
                QMessageBox::Critical,
                QMessageBox::Yes | QMessageBox::Escape,
                QMessageBox::No  | QMessageBox::Default,
                0);

            if (mb.exec() == QMessageBox::No) return;
        }
    }


    delete_tide_record (rec.header.record_number, &db);


    tideEditor->setRedrawZeroLevel ();

    tideEditor->redrawMap ();
}



//  Print station information.

void
TideDialog::slotPrintStation ()
{
    QPrinter *printer = new QPrinter (QPrinter::ScreenResolution);


    printer->setFullPage (TRUE);
    printer->setPageSize (QPrinter::Letter);

    if (printer->setup (tideDialog))
      {
        Q3TextEdit *e = new Q3TextEdit (tideDialog, "editor");
        QString *s = new QString ();
        char string[MONOLOGUE_LENGTH + 256];

        sprintf (string, "\nStation Name: %s\n", rec.header.name);
        s->append (string);
        if (rec.station_id_context[0] != '\0') {
          sprintf (string, "Station ID context: %s\n", rec.station_id_context);
          s->append (string);
        }
        if (rec.station_id[0] != '\0') {
          sprintf (string, "Station ID: %s\n", rec.station_id);
          s->append (string);
        }
	if (rec.date_imported)
	  {
	    sprintf (string, "Date Imported: %d\n", rec.date_imported);
	    s->append (string);
	  }
        sprintf (string, "Latitude: %+.6f\n", rec.header.latitude);
        s->append (string);
        sprintf (string, "Longitude: %+.6f\n", rec.header.longitude);
        s->append (string);
        if (rec.source[0] != '\0') {
          sprintf (string, "Source: %s\n", rec.source);
          s->append (string);
        }
        sprintf (string, "Time Zone: %s\n", get_tzfile (rec.header.tzfile));
        s->append (string);
        sprintf (string, "Country: %s\n", get_country (rec.country));
        s->append (string);
        sprintf (string, "Restriction: %s\n", 
                 get_restriction (rec.restriction));
        s->append (string);
        if (rec.comments[0] != '\0') {
          sprintf (string, "Comments: %s\n", rec.comments);
          s->append (string);
        }
        if (rec.notes[0] != '\0') {
          sprintf (string, "Notes: %s\n", rec.notes);
          s->append (string);
        }
        if (rec.legalese) {
  	  sprintf (string, "Legalese: %s\n", get_legalese (rec.legalese));
	  s->append (string);
        }
        if (rec.xfields[0] != '\0') {
          s->append (rec.xfields);
          if (rec.xfields[strlen(rec.xfields)-1] != '\n')
            s->append ("\n");
        }
	if (rec.min_direction != 361 || rec.max_direction != 361) {
	  sprintf (string, "Direction Units: %s\n", 
		   get_dir_units (rec.direction_units));
	  s->append (string);
        }
	if (rec.min_direction != 361)
	  {
	    sprintf (string, "Ebb Direction: %d\n", rec.min_direction);
	    s->append (string);
	  }
	if (rec.max_direction != 361)
	  {
	    sprintf (string, "Flood Direction: %d\n", rec.max_direction);
	    s->append (string);
	  }
	sprintf (string, "Level Units: %s\n", get_level_units (rec.level_units));
	s->append (string);


        if (rec.header.record_type == REFERENCE_STATION)
          {
            unsigned i;

            sprintf (string, "Datum Description: %s\n", get_datum (rec.datum));
            s->append (string);
            sprintf (string, "Datum Value: %f\n", rec.datum_offset);
            s->append (string);
            sprintf (string, "Meridian (Zone Offset): %d\n", rec.zone_offset);
            s->append (string);
            if (rec.expiration_date)
              {
                sprintf (string, "Expiration Date: %u\n", rec.expiration_date);
                s->append (string);
              }
            if (rec.months_on_station)
              {
                sprintf (string, "Months On Station: %u\n",
                         rec.months_on_station);
                s->append (string);
              }
            if (rec.last_date_on_station)
              {
                sprintf (string, "Last Date On Station: %u\n", 
                         rec.last_date_on_station);
                s->append (string);
              }
            sprintf (string, "Confidence: %u\n", rec.confidence);
            s->append (string);


            sprintf (string, "\n\nConstituents:\n\n");
            s->append (string);

            for (i = 0 ; i < db.constituents ; ++i)
              {
                int j = cnst[i].num;
                if (rec.amplitude[j] == 0.0 && rec.epoch[j] == 0.0)
                  {
                    if (inf_rec.amplitude[j] != 0.0 || inf_rec.epoch[j] != 0.0)
                      {
                        //  Inferred constituents->

                        sprintf (string, 
                            "%s: \tamplitude: %f \tepoch: %f \t(inferred)\n",
                            cnst[i].name, inf_rec.amplitude[j], 
                            inf_rec.epoch[j]);
                        s->append (string);
                      }
                  }
                else
                  {
                    sprintf (string, "%s: \tamplitude: %f \tepoch: %f\n",
                             cnst[i].name, rec.amplitude[j], rec.epoch[j]);
                    s->append (string);
                  }
              }


            TIDE_RECORD temp_rec;
            int subs = 0;
        
            for (i = 0 ; i < db.number_of_records ; ++i)
              {
                get_partial_tide_record (i, &temp_rec.header);

                if (temp_rec.header.reference_station == 
                    rec.header.record_number)
                  {
                    if (!subs)
                      {
                        sprintf (string, "\n\nSubordinate Stations:\n\n");
                        s->append (string);
                      }

                    sprintf (string, "%s\n", temp_rec.header.name);
                    s->append (string);
                    ++subs;
                  }                    
              }
          }
        else
          {
            sprintf (string, "\nReference Station: %s\n", 
                     get_station (rec.header.reference_station));
            s->prepend (string);
            if (rec.min_time_add)
              {
                sprintf (string, "Minimum Time Add: %d\n", rec.min_time_add);
                s->append (string);
              }
            if (rec.min_level_add != 0.00)
              {
                sprintf (string, "Minimum Level Add: %f\n", rec.min_level_add);
                s->append (string);
              }
            if (rec.min_level_multiply != 0.00)
              {
                sprintf (string, "Minimum Level Multiply: %f\n", 
                         rec.min_level_multiply);
                s->append (string);
              }
            if (rec.max_time_add)
              {
                sprintf (string, "Maximum Time Add: %d\n", rec.max_time_add);
                s->append (string);
              }
            if (rec.max_level_add != 0.00)
              {
                sprintf (string, "Maximum Level Add: %f\n", rec.max_level_add);
                s->append (string);
              }
            if (rec.max_level_multiply != 0.00)
              {
                sprintf (string, "Maximum Level Multiply: %f\n", 
                         rec.max_level_multiply);
                s->append (string);
              }
            if (rec.flood_begins != NULLSLACKOFFSET)
              {
                sprintf (string, "Flood Begins: %d\n", rec.flood_begins);
                s->append (string);
              }
            if (rec.ebb_begins != NULLSLACKOFFSET)
              {
                sprintf (string, "Ebb Begins: %d\n", rec.ebb_begins);
                s->append (string);
              }
          }

        e->setText (*s);


        QPainter p;
        if (!p.begin (printer)) return;


        Q3PaintDeviceMetrics metrics (p.device ());
        int dpiy = metrics.logicalDpiY ();
        int margin = (int) ((2 / 2.54) * dpiy);                 // 2 cm margins
        QRect body (margin, margin, metrics.width() - 2 * margin, 
                    metrics.height () - 2 * margin);
        Q3SimpleRichText richText (
            Q3StyleSheet::convertFromPlainText (e->text ()), QFont (),
            e->context (), e->styleSheet (), e->mimeSourceFactory (),
            body.height());
        richText.setWidth (&p, body.width ());
        QRect view (body);
        int page = 1;
        do 
          {
            richText.draw (&p, body.left (), body.top (), view, 
                           colorGroup ());
            view.moveBy (0, body.height ());
            p.translate (0, -body.height ());
            p.drawText (view.right () - 
                        p.fontMetrics ().width (QString::number (page)),
                        view.bottom () + 
                        p.fontMetrics ().ascent () + 5, 
                        QString::number (page));
            if (view.top () >= richText.height ()) break;
            printer->newPage ();
            ++page;
          } while (TRUE);
      } 
}



//  Change the Reference station ID for one or more Subordinate stations.

void 
TideDialog::slotChangeRefStation ()
{
    TIDE_RECORD temp_rec;
    int count = 0;
    QStringList strlist;


    //  Build a list of displayed stations (not all stations) that have
    //  a header.reference_station value of -1 (true Reference stations not
    //  Subordinate stations with a set of constituents).

    for (int i = 0 ; i < tideEditor->number_of_records ; ++i)
    {
        get_partial_tide_record (tideEditor->station_list[i].recnum, 
            &temp_rec.header);
        if (temp_rec.header.reference_station == -1)
        {
            strlist += tideEditor->station_list[i].name;
            ++count;
        }
    }


    //  If there were any in the current displayed area pop up the input
    //  dialog and get the selected station.

    if (count)
    {
        bool ok = FALSE;
        QString res = QInputDialog::getItem (tr ("Reference Stations"),
            tr ("Please select the new reference station"), strlist, 0, FALSE,
            &ok, (QWidget *) this);

        if (ok)
        {
            char station[ONELINER_LENGTH];
            safeStrCpy (station, res, ONELINER_LENGTH);

            int num = find_station (station);
            for (unsigned i = 0 ; i < subList->count () ; ++i)
            {
                if (subList->isSelected (i))
                {
                    read_tide_record (sub_station[i], &temp_rec);
                    temp_rec.header.reference_station = num;
                    update_tide_record (sub_station[i], &temp_rec, &db);
                }
            }
        }
    }
}
